local _,stash = ...
local L = stash.localization
local C = stash.constants
stash.tooltips = {}

local gameTooltip = StashGameTooltip
local scanTooltip = StashGameTooltipScan
local moneyTooltip = StashMoneyTooltip

local suppressTooltip = false

function stash.tooltips:Setup()
    stash.tooltips:UpdateTooltipHooks() -- doesn't actually if AddToTooltip setting disabled
    stash.tooltips:HookItemTooltip(gameTooltip,true) -- StashGameTooltip always hooked
end

-- creates an OnTooltipSetItem hook for the given tooltip
function stash.tooltips:HookItemTooltip(tooltip,always)
    if C.CLIENT_BUILD>=100002 then -- if on 10.0.2 or later (this should be called only once ever)
        TooltipDataProcessor.AddTooltipPostCall(Enum.TooltipDataType.Item, function(tooltip,data)
            --[[ WARNING: Enable RetailAddToTooltips at your own risk. This is not supported. ]]
            if (tooltip==gameTooltip or stash.settings.RetailAddToTooltips) and tooltip~=scanTooltip then
                if data.id then -- data.id is the itemID; in theory should always be defined
                    stash.tooltips:AddLocationsToTooltip(tooltip,data.id,false)
                end
            end
        end)
    else
        local lastTimeAdded = nil
        local numAddedThisFrame = nil
        tooltip:HookScript("OnTooltipSetItem",function(self)
            if always or stash.settings.AddToTooltips then
                -- tooltip scans are already computationally expensive, and recursively searching through stash
                -- would make it ridiculous. on the assumption that tooltip scans happen in the same execution
                -- path (or same "frame") and GetTime() won't advance, this adds stash details only on the first
                -- couple attempts per frame. (Using 2 instead of 1 since recipes have an OnTooltipSetItem twice,
                -- and we want to ignore the call on the first one for a recipe.)
                local now = GetTime()
                if lastTimeAdded==now then
                    numAddedThisFrame = numAddedThisFrame + 1
                else
                    numAddedThisFrame = 0
                end
                local _,link = self:GetItem()
                if link then
                    local itemID = tonumber(link:match("item:(%d+)"))
                    if itemID then
                        if numAddedThisFrame < 2 then
                            local _,_,_,_,_,itemClass,itemSubclass = GetItemInfoInstant(itemID)
                            if lastTimeAdded==now or itemClass~=Enum.ItemClass.Recipe or itemSubclass==Enum.ItemRecipeSubclass.Book or itemSubclass==Enum.ItemRecipeSubclass.Enchanting then
                                stash.tooltips:AddLocationsToTooltip(self,itemID,false)
                            end
                        end
                        lastTimeAdded = now
                    end
                end
            end
        end)
    end
end

-- to prevent Stash from touching outside tooltips at all for users who don't have AddToTooltips enabled,
-- only hook if it's ever enabled; this should be called on login and when the setting changes.
-- in 10.0.2 this function never does anything but classic clients still need it
local tooltipsHooked = false
function stash.tooltips:UpdateTooltipHooks()
    if C.CLIENT_BUILD<100002 then -- starting in 10.0.2, the TooltipDataProcessor.AddTooltipPostCall is always active
        if not tooltipsHooked and stash.settings.AddToTooltips then
            -- hook for item tooltips
            stash.tooltips:HookItemTooltip(GameTooltip)
            -- -- currency and pet tooltips don't happen on retail; and presumably aren't tooltip scanned except rarely
            -- if C.IS_MAINLINE then
            --     -- hook for currency tooltips
            --     hooksecurefunc(GameTooltip,"SetCurrencyToken",function(self,index)
            --         if stash.settings.AddToTooltips then
            --             local link = C_CurrencyInfo.GetCurrencyListLink(index)
            --             if link then
            --                 _stash.tooltips:AddLocationsToTooltip(GameTooltip,(link:match("(currency:%d+)")),false)
            --             end
            --         end
            --     end)
            --     -- hook for battle pet tooltips
            --     hooksecurefunc("BattlePetToolTip_Show",function(speciesID)
            --         if stash.settings.AddToTooltips and tonumber(speciesID) then
            --             local locations = _stash.info:GetLocationsByItemID("battlepet:"..speciesID,false)
            --             if #locations>0 then
            --                 for _,line in ipairs(locations) do
            --                     BattlePetTooltipTemplate_AddTextLine(BattlePetTooltip,line,1,1,1,true)
            --                 end
            --             end
            --         end
            --     end)
            -- end
            tooltipsHooked = true
        end
    end
end

-- hides all tooltips
function stash.tooltips:HideTooltip()
    gameTooltip:Hide()
    moneyTooltip:Hide()
end

-- when something is likely to show a tooltip and we don't want it shown (such as changing from moneypanel to itemspanel
-- while mouse was over okay button but is now over moneybutton), call this to prevent it from being shown for a short time
function stash.tooltips:Pause()
    suppressTooltip = true
    stash.timer:Start(C.PAUSE_TOOLTIP_TIMER,stash.tooltips.Resume,stash.tooltips)
end

-- called auto
function stash.tooltips:Resume()
    suppressTooltip = false
end

-- for simple tooltips like the minimap button tooltip
function stash.tooltips:ShowSimpleTooltip(parent,anchor,title,body)
    gameTooltip:SetOwner(parent,anchor or "ANCHOR_LEFT")
    gameTooltip:AddLine(title)
    gameTooltip:AddLine(body)
    gameTooltip:Show()
end

-- called when an item tooltip had no item (wasn't cached) to retry the tooltip after a short delay
function stash.tooltips:RetryItemTooltip(button,itemID)
    self.retryButton = button
    self.retryItemID = itemID
    stash.timer:Start(C.DATA_CACHE_DELAY,self.RetryItemTooltipCallback,self)
end

-- called on a delay from RetryItemTooltip, does a new OnEnter
function stash.tooltips:RetryItemTooltipCallback()
    local focus = GetMouseFocus()
    if focus and focus.itemID and focus.itemID==self.retryItemID and focus==self.retryButton and MouseIsOver(StashFrame) then
        focus:GetScript("OnEnter")(focus)
    end
    self.retryButton = nil
    self.retryItemID = nil
end

-- called in the onenter of an item list button, displays a tooltip of the item and appends who has the item where
function stash.tooltips:ShowItemTooltip(button,itemID)
    if suppressTooltip then
        return
    end
    local sx,sy = button:GetCenter()
    local ss = button:GetEffectiveScale()
    local ux,uy = UIParent:GetCenter()
    local us = UIParent:GetEffectiveScale()
    gameTooltip:SetOwner(button,sx*ss<ux*us and "ANCHOR_RIGHT" or "ANCHOR_LEFT")
    if type(itemID)=="number" then
        gameTooltip:SetItemByID(itemID)
        if not gameTooltip:GetItem() then -- if tooltip item is not cached then try again in a bit
            self:RetryItemTooltip(button,itemID)
        end
    elseif type(itemID)=="string" and itemID:match("^battlepet") then
        self:SetBattlePetTooltip(itemID)
        stash.tooltips:AddLocationsToTooltip(gameTooltip,itemID,false)
    elseif type(itemID)=="string" and itemID:match("^currency") then
        self:SetCurrencyTooltip(itemID)
        stash.tooltips:AddLocationsToTooltip(gameTooltip,itemID,false)
    else
        return -- don't know what this is, don't show a tooltip
    end
    gameTooltip.itemID = itemID
end

function stash.tooltips:ShowMoneyTooltip(button)
    if suppressTooltip then
        return
    end
    local money = stash.info:FindMoney()
    if not moneyTooltip.lines then
        moneyTooltip.lines = {}
    end
    local maxWidth = 0 -- becomes width of widest button
    local numLines = 0 -- number of lines added to tooltip
    local hasMoneyElsewhere = false -- becomes true if there's money in other realms/factions
    -- set all lines
    for i=1,#money do
        -- but only for current realm and faction (rest can go in list)
        if money[i][C.FOUND_REALM]==stash.current.realm and money[i][C.FOUND_FACTION]==stash.current.faction then
            if not moneyTooltip.lines[i] then
                moneyTooltip.lines[i] = CreateFrame("Frame",nil,moneyTooltip,"StashMoneyTooltipLineTemplate")
                moneyTooltip.lines[i]:SetPoint("TOPLEFT",8,-6-(i-1)*18)
            end
            local color = money[i][C.FOUND_PLAYER]:match(" ") and C.HEX_GREEN or C.FACTION_COLORS[money[i][C.FOUND_FACTION]] or C.HEX_NEUTRAL
            moneyTooltip.lines[i].Name:SetText(format("%s%s",color,money[i][C.FOUND_PLAYER]))
            moneyTooltip.lines[i].Money:SetAmount(money[i][C.FOUND_COUNT])
            moneyTooltip.lines[i].Stripe:SetShown(i%2==0)
            maxWidth = max(maxWidth,moneyTooltip.lines[i].Name:GetStringWidth()+moneyTooltip.lines[i].Money:GetWidth()+8)
            numLines = numLines + 1
        else
            hasMoneyElsewhere = true
        end
    end

    -- add a line that says "Click for all realms/factions" if there's money on other realms or factions
    if hasMoneyElsewhere then
        moneyTooltip.Instructions:SetText(format(L["%s Click for all realms/factions"],C.LMB))
        maxWidth = max(maxWidth,moneyTooltip.Instructions:GetStringWidth())
        moneyTooltip.Instructions:Show()
    else
        moneyTooltip.Instructions:Hide()
    end

    -- resize all lines to max width
    for i=1,numLines do
        moneyTooltip.lines[i]:SetWidth(maxWidth)
        moneyTooltip.lines[i]:Show()
    end
    -- hide any leftover lines (characters reset or filters changed)
    for i=numLines+1,#moneyTooltip.lines do
        moneyTooltip.lines[i]:Hide()
    end
    moneyTooltip:SetSize(maxWidth+16,numLines*18+12+(hasMoneyElsewhere and 18 or 0))
    moneyTooltip:SetPoint("BOTTOMRIGHT",button,"TOPRIGHT",0,0)
    moneyTooltip:SetBackdropColor(0.075,0.075,0.075)
    --moneyTooltip:SetBackdropBorderColor(0.5,0.5,0.5)
    moneyTooltip:Show()
    moneyTooltip.numLines = numLines
end

-- for caged pets, not using the default UI's BattlePetTooltip since the stats are unknown. this adds some
-- info to gameTooltip about the caged pet such as name, type, collected count and source
function stash.tooltips:SetBattlePetTooltip(itemID)
    if suppressTooltip then
        return
    end
    local speciesID = tonumber((itemID or ""):match("battlepet:(%d+)"))
    if speciesID then
        local name,icon,petType,_,source,_,_,canBattle = C_PetJournal.GetPetInfoBySpeciesID(speciesID)
        local count,maxCount = C_PetJournal.GetNumCollectedInfo(speciesID)
        if name then
            gameTooltip:AddLine(C.HEX_WHITE..name)
            gameTooltip:AddLine(C.HEX_BLUE..(canBattle and L["Caged Battle Pet"] or L["Caged Companion Pet"]))
            gameTooltip:AddLine(C.HEX_GREEN.._G["BATTLE_PET_NAME_"..petType])
            gameTooltip:AddLine(((not count or count==0) and C.HEX_WHITE or count==maxCount and C.HEX_RED or C.HEX_GREEN)..format(ITEM_PET_KNOWN,count,maxCount))
            gameTooltip:AddLine(" ")
            gameTooltip:AddLine(C.HEX_WHITE..source)
        end
    end
end

-- adds the locations that the itemID in the current gameTooltip can be found to the tooltip
function stash.tooltips:AddLocationsToTooltip(tooltip,itemID,addSpace)
    local locations = stash.info:GetLocationsByItemID(itemID,addSpace)
    local numLocations = #locations
    if numLocations>0 and tooltip~=scanTooltip then
        for _,line in ipairs(locations) do
            tooltip:AddLine(line)
        end
        tooltip:Show()
    end
    return numLocations
end

function stash.tooltips:SetCurrencyTooltip(itemID)
    local currencyID = tonumber(itemID:match("(%d+)"))
    if not currencyID then
        return
    end
    local info = C_CurrencyInfo.GetBasicCurrencyInfo(currencyID)
    if info then
        local color = C.QUALITY_COLORS[info.quality or 1] or C.QUALITY_COLORS[1]
        gameTooltip:AddLine(color.hex..info.name)
        gameTooltip:AddLine(C.HEX_BLUE..L["Currency"])
        gameTooltip:AddLine(info.description,1,0.82,0,true)
        local extraInfo = C_CurrencyInfo.GetCurrencyInfo(currencyID)
        if extraInfo.maxQuantity and extraInfo.maxQuantity>0 then
            gameTooltip:AddLine(" ")
            gameTooltip:AddLine(format(L["Maximum %s%s\124r per character"],C.HEX_WHITE,BreakUpLargeNumbers(extraInfo.maxQuantity)))
        end
    end
end

